# -*- coding: utf-8 -*-

WebAPI = ['dereference', 'exists',
          'getCount', 'getDefinition', 'getHead', 'getId', 'getLastVersion', 'getPlace', 'getPoolname', 'getPoolurl', 'getPoolpath', 'getProductClasses', 'getReferenceCount', 'getScheme', 'getTagUrnMap', 'getTags', 'getUrn', 'getUrnId', 'getUrnObject', 'getVersions',
          'isAlive', 'isEmpty', 'loadDescriptors', 'loadProduct', 'lockpath',
          'meta', 'mfilter',
          'pfilter', 'poolname', 'poolurl', 'readHK', 'reference', 'remove', 'removeAll', 'removeTag', 'removeUrn', 'removekey',
          'saveDescriptors', 'saveProduct', 'saveProductRef', 'schematicLoad', 'schematicRemove', 'schematicSave', 'schematicSelect', 'schematicWipe', 'select', 'setTag', 'setup',
          'tagExists']
